/*-
 * ============LICENSE_START=======================================================
 *  Copyright (C) 2023-2024 Nordix Foundation.
 * ================================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * SPDX-License-Identifier: Apache-2.0
 * ============LICENSE_END=========================================================
 */

package org.onap.policy.clamp.acm.participant.intermediary.handler;

import io.opentelemetry.context.Context;
import java.io.Closeable;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import lombok.RequiredArgsConstructor;
import org.onap.policy.clamp.acm.participant.intermediary.api.AutomationCompositionElementListener;
import org.onap.policy.clamp.acm.participant.intermediary.api.CompositionDto;
import org.onap.policy.clamp.acm.participant.intermediary.api.CompositionElementDto;
import org.onap.policy.clamp.acm.participant.intermediary.api.InstanceElementDto;
import org.onap.policy.clamp.acm.participant.intermediary.api.ParticipantIntermediaryApi;
import org.onap.policy.clamp.models.acm.concepts.AcTypeState;
import org.onap.policy.clamp.models.acm.concepts.DeployState;
import org.onap.policy.clamp.models.acm.concepts.LockState;
import org.onap.policy.clamp.models.acm.concepts.ParticipantRestartAc;
import org.onap.policy.clamp.models.acm.concepts.StateChangeResult;
import org.onap.policy.models.base.PfModelException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

@Component
@RequiredArgsConstructor
public class ThreadHandler implements Closeable {
    private static final Logger LOGGER = LoggerFactory.getLogger(ThreadHandler.class);

    private final AutomationCompositionElementListener listener;
    private final ParticipantIntermediaryApi intermediaryApi;
    private final CacheProvider cacheProvider;

    private final Map<UUID, Future> executionMap = new ConcurrentHashMap<>();

    private final ExecutorService executor =
            Context.taskWrapping(Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors()));

    /**
     * Handle a deploy on a automation composition element.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     */
    public void deploy(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() -> this.deployProcess(compositionElement, instanceElement));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void deployProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        try {
            listener.deploy(compositionElement, instanceElement);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element deploy failed {} {}", instanceElement.elementId(),
                e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElement.instanceId(),
                instanceElement.elementId(), DeployState.UNDEPLOYED, null, StateChangeResult.FAILED,
                "Automation composition element deploy failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    /**
     * Handle an udeploy on a automation composition element.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     */
    public void undeploy(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() -> this.undeployProcess(compositionElement, instanceElement));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void undeployProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        try {
            listener.undeploy(compositionElement, instanceElement);
        } catch (PfModelException e) {
            LOGGER.error(
                "Automation composition element undeploy failed {} {}", instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElement.instanceId(),
                instanceElement.elementId(), DeployState.DEPLOYED, null,
                StateChangeResult.FAILED, "Automation composition element undeploy failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    /**
     * Handle a automation composition element lock.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     */
    public void lock(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() -> this.lockProcess(compositionElement, instanceElement));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void lockProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        try {
            listener.lock(compositionElement, instanceElement);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element lock failed {} {}",
                instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElement.instanceId(),
                instanceElement.elementId(), null, LockState.UNLOCKED, StateChangeResult.FAILED,
                "Automation composition element lock failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    /**
     * Handle a automation composition element unlock.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     */
    public void unlock(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() -> this.unlockProcess(compositionElement, instanceElement));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void unlockProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        try {
            listener.unlock(compositionElement, instanceElement);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element unlock failed {} {}",
                instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElement.instanceId(),
                instanceElement.elementId(), null, LockState.LOCKED, StateChangeResult.FAILED,
                "Automation composition element unlock failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    /**
     * Handle a automation composition element delete.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     */
    public void delete(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() -> this.deleteProcess(compositionElement, instanceElement));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void deleteProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement) {
        try {
            listener.delete(compositionElement, instanceElement);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element delete failed {} {}",
                instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(
                instanceElement.instanceId(), instanceElement.elementId(), DeployState.UNDEPLOYED, null,
                StateChangeResult.FAILED, "Automation composition element delete failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    /**
     * Handle a automation composition element properties update.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param instanceElement the information of the Automation Composition Instance Element
     * @param instanceElementUpdated the information of the Automation Composition Instance Element updated
     */
    public void update(UUID messageId, CompositionElementDto compositionElement, InstanceElementDto instanceElement,
                       InstanceElementDto instanceElementUpdated) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() ->
            this.updateProcess(compositionElement, instanceElement, instanceElementUpdated));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void updateProcess(CompositionElementDto compositionElement, InstanceElementDto instanceElement,
                               InstanceElementDto instanceElementUpdated) {
        try {
            listener.update(compositionElement, instanceElement, instanceElementUpdated);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element update failed {} {}",
                instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElement.instanceId(),
                instanceElement.elementId(), DeployState.DEPLOYED, null,
                StateChangeResult.FAILED, "Automation composition element update failed");
        }
        executionMap.remove(instanceElement.elementId());
    }

    private void cleanExecution(UUID execIdentificationId, UUID messageId) {
        var process = executionMap.get(execIdentificationId);
        if (process != null) {
            if (!process.isDone()) {
                process.cancel(true);
            }
            executionMap.remove(execIdentificationId);
        }
        cacheProvider.getMsgIdentification().put(execIdentificationId, messageId);
    }

    /**
     * Handles prime a Composition Definition.
     *
     * @param messageId the messageId
     * @param composition the composition
     */
    public void prime(UUID messageId, CompositionDto composition) {
        cleanExecution(composition.compositionId(), messageId);
        var result = executor.submit(() -> this.primeProcess(composition));
        executionMap.put(composition.compositionId(), result);
    }

    private void primeProcess(CompositionDto composition) {
        try {
            listener.prime(composition);
            executionMap.remove(composition.compositionId());
        } catch (PfModelException e) {
            LOGGER.error("Composition Defintion prime failed {} {}", composition.compositionId(), e.getMessage());
            intermediaryApi.updateCompositionState(composition.compositionId(), AcTypeState.COMMISSIONED,
                StateChangeResult.FAILED, "Composition Defintion prime failed");
        }
    }

    /**
     * Handles deprime a Composition Definition.
     *
     * @param messageId the messageId
     * @param composition the composition
     */
    public void deprime(UUID messageId, CompositionDto composition) {
        cleanExecution(composition.compositionId(), messageId);
        var result = executor.submit(() -> this.deprimeProcess(composition));
        executionMap.put(composition.compositionId(), result);
    }

    private void deprimeProcess(CompositionDto composition) {
        try {
            listener.deprime(composition);
            executionMap.remove(composition.compositionId());
        } catch (PfModelException e) {
            LOGGER.error("Composition Defintion deprime failed {} {}", composition.compositionId(), e.getMessage());
            intermediaryApi.updateCompositionState(composition.compositionId(), AcTypeState.PRIMED,
                StateChangeResult.FAILED, "Composition Defintion deprime failed");
        }
    }

    /**
     * Handles restarted scenario.
     *
     * @param messageId the messageId
     * @param composition the composition
     * @param state the state of the composition
     * @param automationCompositionList list of ParticipantRestartAc
     */
    public void restarted(UUID messageId, CompositionDto composition,
            AcTypeState state, List<ParticipantRestartAc> automationCompositionList) {
        try {
            listener.handleRestartComposition(composition, state);
        } catch (PfModelException e) {
            LOGGER.error("Composition Defintion restarted failed {} {}", composition.compositionId(), e.getMessage());
            intermediaryApi.updateCompositionState(composition.compositionId(), state, StateChangeResult.FAILED,
                    "Composition Defintion restarted failed");
        }

        for (var automationComposition : automationCompositionList) {
            for (var element : automationComposition.getAcElementList()) {
                var compositionElement = new CompositionElementDto(composition.compositionId(),
                    element.getDefinition(), composition.inPropertiesMap().get(element.getDefinition()),
                    composition.outPropertiesMap().get(element.getDefinition()));
                var instanceElementDto = new InstanceElementDto(automationComposition.getAutomationCompositionId(),
                    element.getId(), element.getToscaServiceTemplateFragment(),
                    element.getProperties(), element.getOutProperties());
                cleanExecution(element.getId(), messageId);
                var result = executor.submit(() ->
                    this.restartedInstanceProcess(compositionElement, instanceElementDto,
                        element.getDeployState(), element.getLockState()));
                executionMap.put(element.getId(), result);
            }
        }
    }

    private void restartedInstanceProcess(CompositionElementDto compositionElement,
        InstanceElementDto instanceElementDto, DeployState deployState, LockState lockState) {
        try {
            listener.handleRestartInstance(compositionElement, instanceElementDto, deployState, lockState);
            executionMap.remove(instanceElementDto.elementId());
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element deploy failed {} {}",
                instanceElementDto.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(instanceElementDto.instanceId(),
                instanceElementDto.elementId(), deployState, lockState, StateChangeResult.FAILED,
                    "Automation composition element restart failed");
        }
    }

    /**
     * Closes this stream and releases any system resources associated
     * with it. If the stream is already closed then invoking this
     * method has no effect.
     *
     * @throws IOException if an I/O error occurs
     */
    @Override
    public void close() throws IOException {
        executor.shutdown();
    }

    /**
     * Handles AutomationComposition Migration.
     *
     * @param messageId the messageId
     * @param compositionElement the information of the Automation Composition Definition Element
     * @param compositionElementTarget the information of the Automation Composition Definition Element Target
     * @param instanceElement the information of the Automation Composition Instance Element
     * @param instanceElementMigrate the information of the Automation Composition Instance Element updated
     */
    public void migrate(UUID messageId, CompositionElementDto compositionElement,
        CompositionElementDto compositionElementTarget, InstanceElementDto instanceElement,
        InstanceElementDto instanceElementMigrate) {
        cleanExecution(instanceElement.elementId(), messageId);
        var result = executor.submit(() ->
            this.migrateProcess(compositionElement, compositionElementTarget, instanceElement, instanceElementMigrate));
        executionMap.put(instanceElement.elementId(), result);
    }

    private void migrateProcess(CompositionElementDto compositionElement,
        CompositionElementDto compositionElementTarget, InstanceElementDto instanceElement,
        InstanceElementDto instanceElementMigrate) {
        try {
            listener.migrate(compositionElement, compositionElementTarget, instanceElement, instanceElementMigrate);
        } catch (PfModelException e) {
            LOGGER.error("Automation composition element migrate failed {} {}",
                instanceElement.elementId(), e.getMessage());
            intermediaryApi.updateAutomationCompositionElementState(
                instanceElement.instanceId(), instanceElement.elementId(), DeployState.DEPLOYED,
                null, StateChangeResult.FAILED, "Automation composition element migrate failed");
        }
        executionMap.remove(instanceElement.elementId());
    }
}
